# tape.tcl
# tape server drive control

# 4 July 1996 SCL
# DONT DISTINGUISH BETWEEN TAPE DEVICE TYPES
# To change back
# Look for the comment above to find the modified parts of the code
# also all cases of:
#                       set drive [first $freelist]
# should be changed to:
#                       set drive $type

#  26 Sept 95 VP  -  TK migration
# 1/2/94 SCL
#  added devstate "moving" in place of eleven for tape handling systems
#
# 5/1/93 SCL
#  add missing devstate "fifteen" ("error" should be 16 was 15)
#  allow CLOSE action from devstate "error"
#
# 30/11/92 SCL
#  allow FREE action if state is MOUNTING
#  a tape was stuck in MOUNTING and could not be freed - button greyed out
#  other state/allowed actions may be desirable
#
# 20/10/92 SCL
# totally revamped:
#  separate drives from streams (now in task tss)
#  available summary statuses follow drive state
#  available commands follow drive state


set debug 0
set started 0

set srv eg
set SRV EG

set frame TAPE
set frametitle "Tape Drive Control"
set icontitle "TapeCntl"

#    maximum characters in critical names
set driveslen 12
set volslen 10
set fileslen 12

#    initial setting for auto associate option
set tsautomode 1

set dsvisible 0
set fsvisible 0
set ttvisible 0
set ssvisible 0
set ttrefresh 1
set forcebuttons 0

set curdrv "TAPE"

set alias ""
set initvols($alias) ""

set volume ""
set oldvol ""
set filename ""
set openmode -1

if {$env(TCLTS) == "orbx"} then {

   set defblen 32768
   set defrlen $defblen
   set fopenoptions [list Labelled Unlabelled Custom]
} else {

   set defrlen 0
   set defblen 0
   set fopenoptions [list Labelled Custom]
}

set blklen [expr $defblen /1024]

set skcount 1
set skmode 0
set skdirctn 0

set defamode 2
set amode $defamode
set amodelist {0 1 2 8 9 10}
set amodeindex [lsearch $amodelist $amode]
set amodenameslist {"Update" "Read" "Write" "RAW (update)" "RAW (read)" "RAW (write)"}

set deflabels 2
set labels $deflabels
set labellist {8 2}
set labelindex [lsearch $labellist $labels]
set labelnameslist {"Unlabelled" "ANSI"}

set rlen $defrlen
set blen $defblen
set smode 1
set smodenameslist {Striped Duplicated}

set defvlabels 2
set vlabels $defvlabels
set vlabellist {8 2}
set vlabelindex [lsearch $vlabellist $vlabels]
set vlabelnameslist {"Unlabelled" "ANSI"}

#    parameters for density

set defdensity 0
set density 0
set densityindex 0


#    DLT device information

#    list of supported density codes
#      0x89  80GB  DLT8000  Type IV tape + compression
#      0x88  40GB  DLT8000  Type IV tape
#      0x85  70GB  DLT7000  Type IV tape + compression
#      0x84  35GB  DLT7000  Type IV tape
#      0x83  40GB  DLT7000  Type IV tape + compression (DLT4000 mode)
#      0x82  20GB  DLT7000  Type IV tape (DLT4000 mode)
#      0x81  30GB  DLT7000  Type IIIXT tape + compression (DLT2000XT mode)
#      0x81  20GB  DLT7000  Type III tape + compression (DLT2000 mode)
#      0x80  15GB  DLT7000  Type IIIXT tape (DLT2000XT mode)
#      0x80  10GB  DLT7000  Type III tape (DLT2000 mode)

#      0x83  40GB  DLT4000  Type IV tape + compression
#      0x82  20GB  DLT4000  Type IV tape
#      0x81  20GB  DLT4000  Type III tape + compression (DLT2000 mode)
#      0x80  10GB  DLT4000  Type III tape (DLT2000 mode)

set densitylist {0 1 2 0x88 0x89 0x85 0x84 0x83 0x82 0x81 0x81 0x80 0x80 0x83 0x82 0x81 0x80}

#    list of corresponding density descriptions
set densitynameslist {"Default" "Low" "High"  \
      "0x89  80GB  DLT8000  Type IV tape + compression"   \
      "0x88  40GB  DLT8000  Type IV tape"   \
      "0x85  70GB  DLT7000  Type IV tape + compression"   \
      "0x84  35GB  DLT7000  Type IV tape"   \
      "0x83  40GB  DLT7000  Type IV tape + compression (DLT4000 mode)"   \
      "0x82  20GB  DLT7000  Type IV tape (DLT4000 mode)"   \
      "0x81  30GB  DLT7000  Type IIIXT tape + compression (DLT2000XT mode)"   \
      "0x81  20GB  DLT7000  Type III tape + compression (DLT2000 mode)"   \
      "0x80  15GB  DLT7000  Type IIIXT tape (DLT2000XT mode)"   \
      "0x80  10GB  DLT7000  Type III tape (DLT2000 mode)"   \
      "0x83  40GB  DLT4000  Type IV tape + compression"   \
      "0x82  20GB  DLT4000  Type IV tape"   \
      "0x81  20GB  DLT4000  Type III tape + compression (DLT2000 mode)"   \
      "0x80  10GB  DLT4000  Type III tape (DLT2000 mode)"}

#    unknown or no tape
set DLTlength(DLT4000,0x00) 0
set DLTlength(DLT7000,0x00) 0
set DLTlength(DLT8000,0x00) 0

#    cleaning tape!!
set DLTlength(DLT4000,0x81) 0
set DLTlength(DLT7000,0x81) 0
set DLTlength(DLT8000,0x81) 0

#    DLTtape I or II!!
set DLTlength(DLT4000,0x82) 0
set DLTlength(DLT7000,0x82) 0
set DLTlength(DLT8000,0x82) 0

#    DLTtape III
set DLTlength(DLT2000,0x83) 10
set DLTlength(DLT4000,0x83) 10
set DLTlength(DLT7000,0x83) 10
set DLTlength(DLT8000,0x83) 10

#    DLTtape IIIxt
set DLTlength(DLT2000,0x84) 15
set DLTlength(DLT4000,0x84) 15
set DLTlength(DLT7000,0x84) 15
set DLTlength(DLT8000,0x84) 15

#    DLTtape IV
set DLTlength(DLT4000,0x85) 20
set DLTlength(DLT7000,0x85) 35
set DLTlength(DLT8000,0x85) 40

set DLTmedia(0x00) "Unknown or media not present"
set DLTmedia(0x81) "Cleaning Tape"
set DLTmedia(0x82) "DLTtape I or DLTtape II"
set DLTmedia(0x83) "DLTtape III"
set DLTmedia(0x84) "DLTtape IIIxt"
set DLTmedia(0x85) "DLTtape IV"

set DLTdensity(0x00) "Unknown or media not present"
set DLTdensity(0x1a) "Type IV tape (DLT4000 mode) 20 GB (without compression)"
set DLTdensity(0x1b) "Type IV tape 35 GB (without compression)"
set DLTdensity(0x19) "Type III tape 10GB / IIIXT tape 15 GB (without compression)"
set DLTdensity(0x41) "Type IV tape 40 GB (without compression)"
set DLTdensity(0x80) "Type III tape 10GB / IIIXT tape 15 GB (without compression)"
set DLTdensity(0x81) "Type III tape 20GB (DLT2000XT mode) / IIIXT tape 30 GB (with compression)"
set DLTdensity(0x82) "Type IV tape (DLT4000 mode) 20 GB (without compression)"
set DLTdensity(0x83) "Type IV tape (DLT4000 mode) 40 GB (with compression)"
set DLTdensity(0x84) "Type IV tape 35 GB (without compression)"
set DLTdensity(0x85) "Type IV tape 70 GB  (with compression)"
set DLTdensity(0x88) "Type IV tape (DLT8000 mode) 40 GB (without compression)"
set DLTdensity(0x89) "Type IV tape (DLT8000 mode) 80 GB  (with compression)"

set automount 1

if {$env(TCLTS) == "orbx"} then {

#
# The Euroball tape server implements various extra states and also
# requires that a header is written after a file is opened
# i.e. the server will not move to state "open" until a header has been written
#

set devstates {unallocated allocated mounting mounted opening open closing executing initialising   \
     identifying putting eleven rewinding thirteen fourteen fifteen "device error" moving skipping \
     importing "writing file mark" test "waiting for header" "writing header"}

set eleven [lsearch $devstates eleven]
set twelve 12
set rewinding [lsearch $devstates rewinding]

set skipping [lsearch $devstates skipping]
set importing [lsearch $devstates importing]
set writingfm [lsearch $devstates "writing file mark"]
set test [lsearch $devstates test]
set waitforhdr [lsearch $devstates "waiting for header"]
set writinghdr [lsearch $devstates "writing header"]

set comment ""
set header ""

} else {

#    RPC server

set devstates {unallocated allocated mounting mounted opening open closing executing initialising \
    identifying putting moving twelve thirteen fourteen fifteen "device error"}

set twelve [lsearch $devstates twelve]
}

set procstatuses { {O.K.} {Client identifier already in use} {Client identifier unknown} {No free resources} \
     {Connect error} {Data management error} {Resource unavailable} {Drive/tape in wrong state} \
     {Invalid capability} {nts_block} {nts_cmd} {Invalid NTS parameter} {nts_nolabel} \
     {EGTS in wrong state} {EGTS already in use} {Invalid EGTS parameter} }

set freelist {}
set usedlist {}
set aliaslist {}
set haltgo {Stop Go  Test}
set nosecs 0
set onesec 1000
set fivesecs 5000
set tensecs 10000
set afterid ""
set tblrows 4
set ttstart 0
set egstatenames {unknown stopped going test}
set smodeslist {S D}


set unallocated [lsearch $devstates unallocated]
set allocated [lsearch $devstates allocated]
set mounting [lsearch $devstates mounting]
set mounted [lsearch $devstates mounted]
set opening [lsearch $devstates opening]
set open [lsearch $devstates open]
set closing [lsearch $devstates closing]
set executing [lsearch $devstates executing]
set initialising [lsearch $devstates initialising]
set identifying [lsearch $devstates identifying]
set putting [lsearch $devstates putting]
set moving [lsearch $devstates moving]
#set twelve [lsearch $devstates twelve]
set thirteen [lsearch $devstates thirteen]
set fourteen [lsearch $devstates fourteen]
set fifteen [lsearch $devstates fifteen]
set errorstate [lsearch $devstates "device error"]

set negstreams 1

set egstate 1
set egstream 1
set egstreamindex 0
set egstreamlist {}

set smode 2
set pluslist {" " "+"}
set plus " "
set seldrstr ""
set usedlist {}
set aliaslist {}
set seldrlist {}
set selidlist {}


if {$env(TCLTS) == "orbx"} then {

#    For Euroball server

proc setup-header {} {
   global header comment
   set title "Euroball Data Tape"
   set date [format "Date: %s" [date]]
   set header [join [list $title "Date: [date]" "Comment: $comment" "\n"] "\n"]
}

proc write-header {} {
        global srv server alias header
        set z [catch {$srv "(tape :path '$server' :alias '$alias' :header '$header' !writeheader)"} m]
        if {$z!=0} then {
                display-$srv-error-report self $m "Attempting to write header $header" "on drive $alias"
                clear-footer
                error $m
        }
}

}

proc claim-drive {d a} {
   global srv server
   set z [catch {$srv "(tape :path '$server' :alias '$a' :drive '$d' !claim)" } m]
   if {$z!=0} then {
                display-$srv-error-report self $m "Attempting to claim drive $d"
                clear-footer
                error $m
                set m ""
   } else {return $m}
}

proc free-drive {} {
        global srv server alias
        set z [catch {$srv "(tape :path '$server' :alias '$alias' !free)" } m]
        if {$z!=0} then {
                display-$srv-error-report self $m "Attempting to free drive $alias"
                clear-footer
                error $m
        }
}

proc initialise-volume {ovol labs dens} {
        global debug srv server alias initvols
        if {$debug} then  {
                puts stdout "initialise-volume: $ovol $labs $dens"
        }
        set z [catch {$srv "(tape :path '$server' :alias '$alias' :volume '$initvols($alias)' :labels $labs :density $dens :oldvol '$ovol' !invol)" } m]
        if {$z!=0} then {
                display-$srv-error-report self $m "Attempting to initialise volume $initvols($alias)"
                clear-footer
                error $m
        }
}

proc identify-volume {} {
        global debug srv server alias
        if {$debug} then  {
                puts stdout "in identify-volume"
        }
        set z [catch {$srv "(tape :path '$server' :alias '$alias' !idvol)" } m]
        if {$z!=0} then {
                display-$srv-error-report self $m "Attempting to identify volume on $alias"
                clear-footer
                error $m
        }
}

proc mount-volume {} {
        global debug srv server alias volume
        if {$debug} then  {
                puts stdout "in mount-volume"
        }
        set z [catch {$srv "(tape :path '$server' :alias '$alias' :volume '$volume' !mntvol)" } m]
        if {$z!=0} then {
                display-$srv-error-report self $m "Attempting to mount volume $volume" "on drive $alias"
                clear-footer
                error $m
        }
}

proc automount-volume al {
        global debug srv server initvols
        if {$debug} then  {
                puts stdout "in automount-volume"
                puts stdout "alias: $al, volume: $initvols($al)"
        }
        set z [catch {$srv "(tape :path '$server' :alias '$al' :volume '$initvols($al)' !mntvol)" } m]
        if {$z!=0} then {
                display-$srv-error-report self $m "Attempting to mount volume $initvols($al)" "on drive $al"

                clear-footer
                error $m
        }
}

proc dismount-volume {} {
        global srv server alias
        set z [catch {$srv "(tape :path '$server' :alias '$alias' !disvol)" } m]
        if {$z!=0} then {
                display-$srv-error-report self $m "Attempting to dismount volume from $alias"
                clear-footer
                error $m
        }
}

proc open-file {mode labs rl bl} {
        global srv server alias filename
        if {$filename!=""} then {
                set fname $filename
        }  else {
                set fname ""
        }
        set z [catch {$srv "(tape :path '$server' :alias '$alias' :amode $mode :labels $labs :rlen $rl :blen $bl :file '$fname' !open)"} m]
        if {$z!=0} then {
                display-$srv-error-report self $m "Attempting to open file $fname" "on drive $alias"
                clear-footer
                error $m
        }
}

proc close-file {} {
        global srv server alias
        set z [catch {$srv "(tape :path '$server' :alias '$alias' !close)" } m]
        if {$z!=0} then {
                display-$srv-error-report self $m "Attempting to close file on drive $alias"
                clear-footer
                error $m
        }
}

proc query-stream-association s {
        global srv server
        set z [catch {$srv "(tape :path '$server' :stream $s ?stream)" } m]
        if {$z!=0} then {
                display-$srv-error-report self $m "Attempting to query stream $s association"
                clear-footer
                error $m
        }
        return $m
}

proc query-file-status al {
        global srv server alias blklen
        set z [catch {$srv "(tape :path '$server' :alias '$al' ?fstatus)" } m]
        if {$z!=0} then {
                if {$m == 0x4000f} then {
                    set alias ""
                } else {
                    display-$srv-error-report self $m "Attempting to query file status" "for drive $al"
                    clear-footer
                    error $m
                }
        } else {
            if {[lindex $m 10] > 0} then {set blklen [expr [lindex $m 10] /1024]}
        }
        return $m
}

proc query-device-status a {
        global env srv server
        global DLTlength DLTInfo

        set z [catch {$srv "(tape :path '$server' :alias '$a' ?dstatus)" } m]
        if {$z!=0} then {
                display-$srv-error-report self $m "Attempting to query device status" "for drive $a"
                clear-footer
                error $m
        }

        set dtype [string range [lindex $m 0] 0 2]
        if {[string toupper $dtype] != "DLT"} then {return $m}

    if {$env(TCLTS) == "orbx"} then {return $m}
    
        if {$env(TCLTS) == "rpc"} then {
	    set DLTsense [lrange $m 9 38]
	    set DLTinquiry [lrange $m 69 124]
	    set DLTmode [lrange $m 45 56]
        } else {
	    set DLTsense [lrange $m 9 38]
	    set DLTinquiry [lrange $m 66 124]
	    set DLTmode [lrange $m 42 53]
        }
       
        if {$env(TCLTS) == "orbx"} then {
	    if {[lindex $DLTinquiry 11] == 34} then {
		set DLTInfo [lreplace $m 0 0 "DLT4000"]
	    }
	    if {[lindex $DLTinquiry 11] == 37} then {
		set DLTInfo [lreplace $m 0 0 "DLT7000"]
	    }
	    if {[lindex $DLTinquiry 11] == 38} then {
		set DLTInfo [lreplace $m 0 0 "DLT8000"]
	    }
	} else {
#	    set DLTInfo $m
	    if {[lindex $DLTinquiry 11] == 34} then {
		set DLTInfo [lreplace $m 0 0 "DLT4000"]
	    }
	    if {[lindex $DLTinquiry 11] == 37} then {
		set DLTInfo [lreplace $m 0 0 "DLT7000"]
	    } 
	    if {[lindex $DLTinquiry 11] == 38} then {
		set DLTInfo [lreplace $m 0 0 "DLT8000"]
	    } 
	}
    
#    device type  DLT4000; DLT7000 etc
        set devicetype [string range [lindex $DLTInfo 0] 0 6]

        if {[llength $DLTInfo] < 38} then {return $m}

#    media type and density code - together these give the tape length
        set mediatype 0x[lindex $DLTmode 1]
        set densitycode 0x[lindex $DLTmode 4]
        if {[info exists DLTlength($devicetype,$mediatype)]} then {
           set tapelength [expr $DLTlength($devicetype,$mediatype) * 1000 *1000]
        } else {
           set tapelength 0
        }
#    tape remaining (Kbytes)
        set tapeleft [expr 0x[lindex $DLTsense 25][lindex $DLTsense 26][lindex $DLTsense 27][lindex $DLTsense 28]*4]
#    tape i/o errors and tape i/o error rate are not currently available
        set DLTInfo [lreplace $DLTInfo 1 4 $tapelength $tapeleft 0 0]

        return $DLTInfo
}

proc find-used-drives {} {
   global debug srv server aliases aliaslist usedlist
   set aliaslist {}
   set usedlist {}
   foreach a $aliases {
      if {$debug} then  {puts stdout "(tape :path '$server' :alias '$a' ?fstatus)"}
      set z [catch {$srv "(tape :path '$server' :alias '$a' ?fstatus)" } m]
      if {$z==0} then {
         lappend aliaslist $a
         set mm [lindex $m 4]
         if {$mm =="SINK"} then {set mm [lindex $m 5]}
         lappend usedlist $mm
      }  else  {
         if {$m != 0x4000f && $m != 5} then {
                display-$srv-error-report self $m "Attempting to find used drives on $server"
                clear-footer
                error $m
         }
      }
   }
}

proc prefix-list {l p} {
   set ll {}
   foreach f $l {lappend ll $p$f}
   return $ll
}

proc read-drives {} {
        global srv server
        set z [catch {$srv "(tape :path '$server' ?drives)"} m]
        if {$z!=0} then {
                display-$srv-error-report self $m "Attempting to read drives for $server"
                clear-footer
                error $m
        }
        return $m
}

proc find-fdrives d {
   global type
   set ll {}
   foreach f $d {if {[third $f] == $type && [first $f] == "0"} then {lappend ll [second $f]}}
   return $ll
}

# DONT DISTINGUISH BETWEEN TAPE DEVICE TYPES
proc find-fdrives d {
   set ll {}
   foreach f $d {if {[first $f] == "0"} then {lappend ll [second $f]}}
   return $ll
   }

proc read-egstate {} {
        global srv server
        set z [catch {$srv "(tape :path '$server' ?state)"} m]
        if {$z!=0} then {
                display-$srv-error-report self $m "Attempting to read state of $server"
                clear-footer
                error $m
        }
        return $m
}

#    following version used by external tasks to determine state of tape subsystem
#    in the event of any error return state=stopped

proc inquire-egstate {} {
        global srv server
        set z [catch {$srv "(tape :path '$server' ?state)"} m]
        if {$z!=0} then {return 0}
        return $m
}

#    Two new procedures added mainly for GANIL and Exogam which use procedures 32 and 33
proc inquire-ganil-status {} {
        global srv server
        set z [catch {$srv "(tape :path '$server' ?gstatus)" } m]
        if {$z!=0} then {
                display-$srv-error-report self $m "Attempting to obtain GANIL Tape Server status"
                clear-footer
                error $m
        }
        return $m
}

proc set-run-number {v} {
        global srv server
        set z [catch {$srv "(tape :path '$server' :run $v !run)" } m]
        if {$z!=0} then {
                display-$srv-error-report self $m "Attempting to set run number for GANIL Tape Server"
                clear-footer
                error $m
        }
}

proc query-egstream-state s {
        global srv server
        set z [catch {$srv "(tape :path '$server' :stream $s ?strstate)"} m]
        if {$z!=0} then {
                display-$srv-error-report self $m "Attempting to read state of stream $s"
                clear-footer
                error $m
        }
        set ll {}
        lappend ll [second $m]
        lappend ll [third $m]
        lappend ll [fourth $m]
        return $ll
}

proc set-stream-association {s idlist} {
        global SRV srv server smode
        set i 0
        set z [catch {$srv "(tape !clear)" } m]
        foreach f $idlist {
                set z [catch {$srv "(tape :idnum $i :assid '$f')" } m]
                incr i
        }

        $SRV "tape :path '$server' :stream $s :smode $smode !stream"
        catch {inform sys redisplay-tapes}
}

proc find-assoc-strings s {
   set qa [query-stream-association $s]
   return [strings-from-ids [lrange $qa 1 end]]
}

proc get-assoc-list s {
   set seldrlist {}
   foreach d $seldrstr {set seldrlist [concat $seldrlist [split $d +]]}
}

proc set-egstate {} {
        global SRV srv server egstate
        catch {$SRV "tape :path '$server' :state $egstate !state"}
        catch {inform sys redisplay-tapes}
}

proc drives-from-aliases al {
   global aliaslist usedlist
   set ll {}
   foreach a $al  {
      set i [lsearch $aliaslist $a]
      if {$i==-1} then {
         lappend ll ????
      } else {
         lappend ll [lindex $usedlist $i]
      }
   }
   return $ll
   }

proc aliases-from-drives dr {
   global aliaslist usedlist
   set ll {}
   foreach d $dr  {
      set i [lsearch $usedlist $d]
      if {$i==-1} then {
         lappend ll ????
      } else {
         lappend ll [lindex $aliaslist $i]
      }
   }
   return $ll
   }

proc strings-from-ids ids {
   set strlist {}
   foreach a $ids {lappend strlist [join [drives-from-aliases [split $a +]] +]}
   return $strlist
}

proc ids-from-strings str {
   set idlist {}
   foreach d $str {lappend idlist [join [aliases-from-drives [split $d +]] +]}
   return $idlist
}

proc ssrepaint {} {
  global env negstreams units blklen
   for {set i 1}  {$i<=$negstreams} {incr i}  {
      set st [query-egstream-state $i]
      set qa [query-stream-association $i]
      set assoclist [drives-from-aliases [lrange $qa 1 end]]
      set widget [format "SS-BLOCK%d" $i]
        :modify $widget :value [first $st]
      set widget [format "SS-BYTE%d" $i]
#        :modify $widget :value [expr [second $st]/$units]
        :modify $widget :value [expr [first $st] * $blklen]
      set widget [format "SS-RATE%d" $i]
        :modify $widget :value [expr [third $st]/$units]
   }
}

proc ssredisplay {} {global ssvisible; if {$ssvisible != 0} then {ssrepaint}}
proc fsredisplay {} {global fsvisible; if {$fsvisible != 0} then {fsrepaint}}
proc dsredisplay {} {global dsvisible; if {$dsvisible != 0} then {dsrepaint}}

proc redisplay {} {
   repaint
   fsredisplay
   dsredisplay
}

proc redisplayall {} {
   redisplay
   drivesrepaint
   ttredisplay
   tsredisplay
   ssredisplay
}

proc tsredisplay {} {
   stateredisplay
   assocsredisplay
}

proc stateredisplay {} {
  global egstate
  set egstate [read-egstate]
  paint-egstate
}

proc ttredisplay {} {
   global debug ttvisible
   if {$debug} then {puts stdout "in ttredisplay"}
   if {$ttvisible != 0} then {
      find-used-drives
      ttrepaint
   }
}

proc ttredo  {} {
   global env debug onesec tensecs afterid ttvisible ttrefresh
   global ttframe
   if {$ttrefresh==0} then {
      if {$debug>1} then {puts stderr "Status refresh is off, kill afterid, don't refresh."}
      if {$afterid!=""} then {kill-after $afterid}
      set afterid ""
      return
   }
   if {$debug>1} then {puts stderr "refreshing status page"}
   if {$ttvisible==1} then {
      set-footer $ttframe "Refreshing display..."
      :modify $ttframe :busy T
      ttredisplay
      set afterid [after $tensecs ttredo]
      :modify $ttframe :busy F
      clear-footer $ttframe
   }
}

proc list-take {as is} {
   set ll {}
   foreach e $is {lappend ll [lindex $as $e]}
   return $ll
}

proc check-state {al fr to ft} {
   global env fivesecs debug srv initialising automount initvols
   global negstreams tsautomode egstate egstream seldrstr plus
   global opening open closing mounted

   if {$env(TCLTS) == "orbx"} then {global waitforhdr writinghdr}

   if {$debug} then {puts stdout "check-state: $al $fr $to $ft"}
#insert-debug  "check-state: $al $fr $to $ft"
#insert-log  "tape: check-state: $al $fr $to $ft"

   set st [query-file-status $al]
   set ps [first $st]
   set ss [second $st]
   set info "[third $st]"
#insert-debug  "check-state: $st"
#insert-log "tape: check-state: $st"

   if {$ss==$fr} then {
#     state is still the "from" state -  so wait a bit longer
      set ftt [format "\"%s\"" $ft]
      if {$debug} then {puts stdout "checking again: $al $fr $to $ftt"}
#insert-debug "checking again: $al $fr $to $ftt"
#insert-log "tape: checking again: $al $fr $to $ftt"
      after $fivesecs check-state $al $fr $to $ftt
      return
   }

   if {$env(TCLTS) == "rpc"} then {
#    rpc
      set failed [expr $ps != 0 && $ps != -1]
#      if {$ps == 0xf070001 && $info == "Operation Successfully Completed"} then {set failed 0}
#      if {$ps == 0xf090001 && $info == "Operation Successfully Completed"} then {set failed 0}
      if {$ps == 0xf070001 || $ps == 0xf090001} then {set failed 0}
      set m [format "0x4000%x" $ps]
      set info "[third $st]\nSee Exception Monitor for more information"
      if {$failed} then {
         redisplayall
         display-$srv-error-report self $m "Found while checking state of $al" "\n$info"
         set-footer [third $st]
         error $m
      }
   } else {
#    Orbix
      set success [expr $ps & 1]
      if {$success == 0 && $info != "command successfully completed"} then {
         redisplayall
         display-$srv-error-report self $ps "Found while checking state of $al" "\n$info"
         set-footer [third $st]
         error 15
      }
   }

   if {$ss==$to} then {
#    now in required state "to" state  -  so operation completed OK
      set-footer $ft
      insert-log "tape: [date] $ft"
      if {$fr==$initialising && $automount==1} then {
         if {$debug} then {puts stdout "automounting $initvols($al) on $al"}
#insert-debug "automounting $initvols($al) on $al"
#insert-log "tape: automounting $initvols($al) on $al"
         do-automount $al
      }

      if {$env(TCLTS) == "orbx"} then {

         if {$fr==$opening && $to==$waitforhdr} then {
            if {$debug} then {puts stdout "writing file header on $al"}
#insert-debug "writing file header on $al"
#insert-log "tape: writing file header on $al"
            do-header
         }
         set frstate $writinghdr
      } else {
         set frstate $opening
      }

      if {$fr==$frstate && $to==$open && $negstreams==1 && $tsautomode==1} then {
         set egstate [read-egstate]

         insert-log "tape: [date] Volume on [drives-from-aliases $al] is [lindex $st 5], file is [first [split [lindex $st 6] /]]"

         if {$egstate == 1 && $egstream == 1} then {
#    file just opened; there is 1 data stream; associate automode is on and
#    the tape server state is STOP and data stream 1 is selected
            set seldrstr [find-assoc-strings 1]
            set drive [drives-from-aliases $al]
            if {[string first $drive $seldrstr] < 0} then {
#    drive not already in association list
               if {$seldrstr == ""} then {set seldrstr $drive} else {append seldrstr "$plus$drive"}
               click-assoc
            }
         }
      }

      if {$fr==$closing && $to==$mounted && $negstreams==1 && $tsautomode==1} then {
         set egstate [read-egstate]
         if {$egstate == 1 && $egstream == 1} then {
#    file just closed; there is 1 data stream; associate automode is on and
#    the tape server state is STOP and data stream 1 is selected
            set seldrstr [find-assoc-strings 1]
            set drive [drives-from-aliases $al]
            set i [lsearch $seldrstr $drive]
            if {$i >= 0} then {
#    remove drive from association list
               set seldrstr [lreplace $seldrstr $i $i]
               click-assoc
            }
         }
      }
   } else {
#    state not in "from" state or "to" state - so must be an error
      set-footer "operation has been aborted due to an error"
      if {$env(TCLTS) == "orbix"} then {
         midas-warning "Drive has not moved to required state\n\
                     A device error has probably occurred\n"
      } else {
         midas-warning "Drive has not moved to required state\n\
                     A device error has probably occurred\n\
                     See Exception Monitor for more information"
      }
   }
   redisplayall
}

proc paint-egstate {} {
   global egstate
     :modify EGSTATE  :value [expr $egstate-1]
}

proc TS-repaint {} {
   global egstreamindex plus pluslist
     :modify EGSTREAM :value $egstreamindex
     :modify PLUS   :value [lsearch $pluslist $plus]
}

proc stringsrepaint {} {
   global seldrstr
   :modify STRINGS :value "$seldrstr"
}

proc assocsredisplay {} {
  global negstreams tsautomode
   for {set i 1} {$i<=$negstreams} {incr i} {
      set widget [format "ASSOCS%d" $i]
        :modify $widget :value "[find-assoc-strings $i]"
   }
  :modify AUTOMODE :value $tsautomode
}

################################################################################


proc click-null {args} {}

proc click-egstream {v args} {
   global egstream egstreamindex egstreamlist
   set egstreamindex  $v
   set egstream [lindex $egstreamlist $v]
   set-footer "Stream $egstream chosen"
}

proc click-strings {s args} {
   global seldrlist seldrstr
   set seldrstr $s
   set seldrlist {}
   foreach d $seldrstr {set seldrlist [concat $seldrlist [split $d +]]}
   drivesrepaint
}

proc click-egstate {v args} {
   global egstate
   set egstates [list invalid stopped going]
   set egstate [expr {$v + 1}]
   set-egstate
   set egstate [read-egstate]
   paint-egstate
   insert-log "tape: [date] server state now [lindex $egstates $egstate]"
   set-footer "Server state set"
}

proc click-assoc {args} {
   global seldrstr egstream egstate
   set egstate [read-egstate]
   if {$egstate == 2} then {
       midas-warning "You cannot change stream association while GOING.\nPlease STOP the server first"
       return
   }
   if {$seldrstr==""} then {click-dissoc; return}
   set-stream-association $egstream [ids-from-strings $seldrstr]
   assocsredisplay
   ssredisplay
   ttredisplay
   insert-log "tape: [date] stream $egstream associated with drive(s) $seldrstr"
   set-footer "Stream $egstream associated with drive(s) $seldrstr"
}

proc click-dissoc {args} {
   global egstream egstate seldrstr seldrlist
   set egstate [read-egstate]
   if {$egstate == 2} then {
       midas-warning "You cannot change stream association while GOING.\nPlease STOP the server first"
       return
   }
   set-stream-association $egstream {}
   assocsredisplay
   ssredisplay
   ttredisplay
   insert-log "tape: [date] stream $egstream disassociated from all drives"
   set-footer "Stream $egstream disassociated from all drives"
}

proc click-clear {args} {
   global seldrlist seldrstr
   set seldrstr ""
   set seldrlist {}
   drivesrepaint
   stringsrepaint
}

proc click-plus {v args} {
   global plus pluslist
   set plus [lindex $pluslist $v]
}

proc click-tsudlist {c d} {
   global env usedlist seldrlist seldrstr plus
   lappend seldrlist $d
   if {$seldrstr!=""} then {
      set seldrstr $seldrstr$plus$d
   } else {
      set seldrstr $d
   }
   set seldrstr [string trim $seldrstr " +"]
   stringsrepaint
   :modify TSUDLIST :select -1
}

proc click-numstrms {v args} {
   global negstreams
   set negstreams $v
   make_tpframe
   redisplayall
}

proc read-numstreams {} {
  global negstreams
  return $negstreams
}

proc click-statuses {} {
   global ssvisible ssframe
   if {$ssvisible==0} then {
      make-ssframe
      set ssvisible 1
   } else {
      open-frame $ssframe
   }
   ssredisplay
   set-footer "Stream statuses queried"
}

proc click-force {v} {
   global forcebuttons
   set forcebuttons $v
   repaint
}

proc click-type v {
   global type typeindex typelist freelist drive curdrv alias
   set typeindex  $v
   set type [lindex $typelist $v]
   set drive [first $freelist]; set curdrv ""; set alias ""
   redisplayall
   set-footer "drive type set"
}

proc click-initvol {v args} {
   global initvols alias
   set initvols($alias) [string toupper $v]
   :modify INITVOL :value "$initvols($alias)"
   set-footer  "Volume name entered"
}

proc click-volume {v args} {
   global volume 
   set volume [string toupper $v]
   :modify VOLUME :value "$volume"
   set-footer  "Volume name entered"
}

proc click-file {v args} {
   global filename
   set filename [string toupper $v]
   :modify FILE :value "$filename"
   set-footer  "File named"
}

proc click-dopopup {mode} {
   global openmode
   set openmode -1
   make_dopanel $mode
   set openmode $mode
   set-footer "Open"
}

proc click-copopup {} {
   global openmode
   make_copanel
   set openmode -1
   set-footer "Custom open"
}

proc click-claim {args} {
   global drive alias aliases aliaslist usedlist curdrv
   global volume filename
   find-used-drives
   foreach a $aliases {
      if {[lsearch $aliaslist $a]==-1} then {set alias $a; break}
   }
   set m [claim-drive $drive $alias]
   set mm $m
   lappend aliaslist $alias
   if {$m =="SINK"} then {
      set m [split $alias P]
      if {[first $m] == "TA"} then {set m SINK[second $m]} else {set m SINK[first $m]}
   }
   lappend usedlist $m
   set curdrv $m; set drive $m
   catch {inform tapechgr update-drivelist}
   set-footer "Claimed $m"
   insert-log "tape: [date] drive $m claimed"
   set egstate [read-egstate]
   if {$mm =="SINK" || [string compare -length 9 $mm "/dev/null"] == 0} then {
       if {$egstate == 1} then {
          set volume $m
          mount-volume
          set volume ""
          set-footer "volume mounted"
          set filename $m
          click-lblopen
          set filename ""
       }
   }
   redisplayall
}

proc click-free {args} {
   global alias aliaslist usedlist freelist drive curdrv type initvols
   set dr $curdrv
   free-drive
   set i [lsearch $aliaslist $alias]
   if {$i!=-1} then {
      set aliaslist [lreplace $aliaslist $i $i]
      set usedlist [lreplace $usedlist $i $i]
   }
   set initvols($alias) ""
   set curdrv ""
   set drive [first $freelist]
   set alias ""
   redisplayall
   catch {inform tapechgr update-drivelist}
   set-footer "Freed $dr"
   insert-log "tape: [date] drive $dr freed"
}

proc click-init {m args} {
   if {$m=="New"}          then {make_nipanel}
   if {$m=="Reinitialise"} then {make_ripanel}
   if {$m=="Custom"}       then {make_cipanel}
}

proc click-mmount {m args} {
   if {$m=="Identify"} then {click-identify}
   if {$m=="Mount"}    then {make_mnpanel}
}

proc click-automount {v args} {
   global automount
   set automount $v
}

proc click-cstinit {args} {
   global oldvol vlabels density
   set-footer "Initialising volume..."
   initialise $oldvol $vlabels $density
   redisplayall
}

proc click-newinit {args} {
   global defvlabels defdensity
   set-footer "Initialising new volume..."
   initialise "" $defvlabels $defdensity
   redisplayall
}

proc click-reinit {args} {
   global oldvol defvlabels defdensity
   set-footer "Reinitialising volume..."
   initialise $oldvol $defvlabels $defdensity
   redisplayall
}

proc initialise {ovol labs dens} {
   global debug nosecs alias initialising allocated initvols curdrv
   if {$debug} then {puts stdout "initialise: $ovol $labs $dens"}
   initialise-volume $ovol $labs $dens
   set ft {"tape on drive $curdrv initialised as volume $initvols($alias)"}
   set-footer "initialising tape on drive $curdrv as volume $initvols($alias) using density code $dens"
   if {$debug} then {puts stdout "checking: $alias $initialising $allocated $ft"}
   after $nosecs check-state $alias $initialising $allocated $ft
}

proc click-vlabel {v args} {
   global vlabelindex vlabels vlabellist ivframe
   set vlabelindex $v
   set vlabels [lindex $vlabellist $v]
   set-footer "labelling chosen"
}

proc click-density {v args} {
   global densityindex density densitylist defdensity ivframe
   set densityindex $v
   set density [lindex $densitylist $v]
   if {$density > 128} then {set density [expr $density + 256]}
   set defdensity $density
   set-footer "density chosen and set as default"
}

proc click-oldvol {v args} {
   global oldvol 
   set oldvol [string toupper $v]
   :modify OLDVOL :value "$oldvol"
   set-footer "old volume name entered"
}

proc click-cireset {args} {
   global defdensity density densityindex defvlabels vlabels vlabelindex
   global volume oldvol automount 
   set density $defdensity
   set densityindex 0
   set vlabels $defvlabels
   set vlabelindex 1
   set oldvol ""
   set automount 1
   make_cipanel
   set-footer  "Initialise defaults reset"
}

proc click-identify {} {
   global debug nosecs alias devstates identifying mounted curdrv
   set-footer "Identifying volume..."
   identify-volume
   set ft {"volume identified and mounted on drive $curdrv"}
   if {$debug} then {puts stdout "checking: $alias $identifying $mounted $ft"}
   after $nosecs check-state $alias $identifying $mounted $ft
   redisplayall
}

proc click-mount {args} {
   global nosecs debug alias mounting mounted curdrv
   set-footer "Mounting volume..."
   mount-volume
   set ft {"volume mounted on drive $curdrv"}
   if {$debug} then {puts stdout "checking: $alias $mounting $mounted $ft"}
   after $nosecs check-state $alias $mounting $mounted $ft
   redisplayall
}

proc do-automount al {
   global nosecs debug mounting mounted curdrv
   set-footer "Automounting volume..."
   automount-volume $al
   set ft {"volume mounted on drive $curdrv"}
   if {$debug} then {puts stdout "checking: $al $mounting $mounted $ft"}
   after $nosecs check-state $al $mounting $mounted $ft
   redisplayall
}

proc click-dismount {args} {
   global volume
   dismount-volume
   redisplayall
   set-footer "Volume $volume dismounted"
}

proc click-mopen {m args} {
   if {$m=="Labelled"}   then {click-dopopup 0}
   if {$m=="Unlabelled"} then {click-unlblopen}
   if {$m=="Custom"}     then {click-copopup}
}

proc click-rewind {args} {
#    rewind device  -  unlabelled (RAW) access only
   global srv server alias
   set z [catch {$srv "(tape :path '$server' :alias '$alias' !rewind)" } m]
   if {$z!=0} then {
      display-$srv-error-report self $m "Attempting to rewind tape."
      clear-footer
      error $m
   }
}

proc click-writetm {args} {
#    write tape mark(s)  -  unlabelled (RAW) access only
   global srv server alias

   set z [catch {$srv "(tape :path '$server' :alias '$alias' !writemark)" } m]
   if {$z!=0} then {
      display-$srv-error-report self $m "Attempting to write tape mark."
      clear-footer
      error $m
   }

}

proc click-skip {args} {
   global srv server alias skcount skmode skdirctn
#    actually skip
#  skdirctn =0  backwards;  skdirctn=1  forwards
#  skmode =0   tape marks;  skmode=1   data blocks

  if {$skdirctn == 0} then {set direction "backwards"} \
    else {set direction "forwards"}
  if {$skmode == 0} then {set mode "tape mark(s)"} \
    else {set mode "data block(s)"}

   set-footer "Skipping $skcount $mode $direction"

   set z [catch {$srv "(tape :path '$server' :alias '$alias' :skmode $skmode :skdir $skdirctn :sknum $skcount !skip)" } m]
   if {$z!=0} then {
      display-$srv-error-report self $m "Attempting to skip."
      clear-footer
      error $m
   }

  clear-footer
}

proc click-skmode {v args} {global skmode; set skmode $v}

proc click-skdirctn {v args} {global skdirctn; set skdirctn $v}

proc click-skcount {v args} {global skcount; set skcount $v}

if {$env(TCLTS) == "orbx"} then {

proc click-custopen {args} {
   global debug nosecs alias opening waitforhdr writinghdr open \
          amode labels rlen blen filename curdrv
   set-footer "Opening file..."
   open-file $amode $labels $rlen $blen
   set ft {"file $filename opened on drive $curdrv"}
   if {$debug} then {puts stdout "checking: $alias $opening $open $ft"}
   if {$amode == 1 || $amode == 9} then {
      after $nosecs check-state $alias $opening $open $ft
   } else {
      after $nosecs check-state $alias $opening $waitforhdr $ft
   }
   redisplayall
}

proc click-lblopen {} {
   global debug nosecs alias opening waitforhdr writinghdr open \
          defamode deflabels defrlen defblen curdrv filename amode
   set-footer "Opening file..."
   open-file $defamode $deflabels $defrlen $defblen
   set ft {"file $filename opened on drive $curdrv"}
   if {$debug} then {puts stdout "checking: $alias $opening $open $ft"}
   if {$amode == 1 || $amode == 9} then {
      after $nosecs check-state $alias $opening $open $ft
   } else {
      after $nosecs check-state $alias $opening $waitforhdr $ft
   }
   redisplayall
}

proc do-header {} {
   global debug nosecs alias opening waitforhdr writinghdr open curdrv
   setup-header
   set-footer "Writing header..."
   write-header
   set ft {"file header written on drive $curdrv"}
   if {$debug} then {puts stdout "checking: $alias $writinghdr $open $ft"}
   after $nosecs check-state $alias $writinghdr $open $ft
   redisplayall
}

} else {

#   rpc

proc click-custopen {args} {
   global debug nosecs alias opening open amode labels rlen blen filename curdrv
   set-footer "Opening file..."
   open-file $amode $labels $rlen $blen
   set ft {"file $filename opened on drive $curdrv"}
   if {$debug} then {puts stdout "checking: $alias $opening $open $ft"}
   after $nosecs check-state $alias $opening $open $ft
   redisplayall
}

proc click-lblopen {} {
   global debug nosecs alias opening open defamode deflabels defrlen defblen filename curdrv
   set-footer "Opening file..."
   open-file $defamode $deflabels $defrlen $defblen
   set ft {"file $filename opened on drive $curdrv"}
   if {$debug} then {puts stdout "checking: $alias $opening $open $ft"}
   after $nosecs check-state $alias $opening $open $ft
   redisplayall
}

}

proc click-unlblopen {} {
   global amode labels rlen blen
   set amode 8
   set labels 8
   click-custopen
}

proc click-close {args} {
   global debug nosecs alias closing mounted curdrv

   set blkcount [lindex [query-file-status $alias] 15]

   set-footer "Closing file..."
   close-file
   if {[string compare -length 4 $curdrv "SINK"] == 0 || [string compare -length 9 $curdrv "/dev/null"] == 0} then {
      set-footer "File closed"
      insert-log "tape: [date] file closed on drive $curdrv after $blkcount blocks"
      set egstate [read-egstate]
      if {$egstate == 1} then {
         click-dissoc
         click-dismount
         click-free
      }
   } else {
      set ft "\"file closed on drive $curdrv after $blkcount blocks\""
      if {$debug} then {puts stdout "checking: $alias $closing $mounted $ft"}
      after $nosecs check-state $alias $closing $mounted $ft
   }
   redisplayall
}

proc click-fdlist {c d} {
   global env drive curdrv type alias usedlist freelist

   :modify UDLIST :deselect [lsearch $usedlist $curdrv]

   switch $c {
     :list-select  {
        if {$d=="TAPE"} then  {
               set drive [first $freelist]; set curdrv ""
            } else {
               set drive $d; set curdrv $d
            }
         }
     :list-deselect  {
        if {$d == $curdrv} then  {
            set drive [first $freelist]; set curdrv ""
            }
        }
     }
   set alias ""

   redisplay
}

proc click-udlist {c d} {
   global env type alias aliaslist drive curdrv usedlist freelist

   :modify FDLIST :deselect [lsearch $freelist $curdrv]

   switch $c {
     :list-select {
         set i [lsearch $usedlist $d]
         set curdrv $d; set drive $d; set alias [lindex $aliaslist $i]
         }
     :list-deselect {
         if {$d==$curdrv} then  {
            set drive [first $freelist]; set curdrv ""; set alias ""
            }
         }
      }
   redisplay
}

proc click-changer {args} {inform tapechgr start}

proc click-status {m args} {
   if {$m=="Summary"} then {click-summary}
   if {$m=="File"}    then {click-status-file}
   if {$m=="Drive"}   then {click-status-device}
   if {$m=="Streams"} then {click-statuses}
}

proc click-summary {} {
   global started tensecs afterid ttvisible ttrefresh
   if {$ttvisible==0} then {
      if {$started != 0} then {set-footer "Starting tape status (all drives)"}
      make_ttframe
      set ttvisible 1
      set ttrefresh 1
   } else  {
      global ttframe; open-frame $ttframe
   }
   if {$started == 0} then  {close-frame}
   if {$afterid!=""} then {kill-after $afterid}
   set afterid ""
   set afterid [after $tensecs ttredo]
   ttredisplay
   if {$started != 0} then {set-footer "Started tape status (all drives)"}
}

proc click-status-file {} {
   global fsvisible
   if {$fsvisible==0} then {
      make_fsframe
      set fsvisible 1
   } else {
      global fsframe; open-frame $fsframe
   }
   make_fsframe
   fsredisplay
   set-footer "Tape status queried (Selected drive)"
}

proc click-status-device {} {
   global dsvisible
   if {$dsvisible==0} then {
      make_dsframe
      set dsvisible 1
   } else {
      global dsframe; open-frame $dsframe
   }
   dsredisplay
   set-footer "Drive status queried"
}

proc click-amode {v args} {
   global amodeindex amode amodelist
   set amodeindex $v
   set amode [lindex $amodelist $v]
   set-footer  "Access mode modified"
}

proc click-label {v args} {
   global labelindex labels labellist
   set labelindex $v
   set labels [lindex $labellist $v]
   set-footer  "Labelling chosen"
}

proc click-rlen {v args}  {
   global rlen blen 
   set rlen $v
   if {$rlen > $blen} then {set rlen $blen}
   :modify RLEN :value $rlen
   set-footer  "Record length altered"
}

proc click-blen {v args}  {
   global rlen blen 
   set blen $v
   if {$rlen > $blen} then {set rlen $blen}
   :modify RLEN :value $rlen
   set-footer  "Block length altered"
}

proc click-opreset {args} {
   global amode amodeindex labels labelindex rlen blen 
   global defamode deflabels defrlen defblen labellist
   set amode $defamode
   set amodeindex 2
   set labels $deflabels
   set labelindex [lsearch $labellist $labels]
   set rlen 0
   set blen 0
   make_copanel
   set-footer  "Open defaults reset"
}

proc click-ttdrives {w args} {
   if {$w=="Start of List"} then {click-ttrestart}
   if {$w=="Next"}          then {click-ttnext1}
   if {$w=="Previous"}      then {click-ttprev1}
   if {$w=="Next 4"}        then {click-ttnext4}
   if {$w=="Previous 4"}    then {click-ttprev4}
}

proc click-ttnext1 {} {
   global ttstart aliaslist
   if {[expr {$ttstart + 4}] < [llength $aliaslist]} then  {
      set ttstart [expr {$ttstart + 1}]
   }
   ttredisplay
}

proc click-ttprev1 {} {
   global ttstart
   set ttstart [expr {$ttstart - 1}]
   if {$ttstart < 0} then  {
      set ttstart 0
   }
   ttredisplay
}

proc click-ttnext4 {} {
   global ttstart aliaslist
   if {[expr {$ttstart + 4}] < [llength $aliaslist]} then  {
      set ttstart [expr {$ttstart + 4}]
   }
   ttredisplay
}

proc click-ttprev4 {} {
   global ttstart
   set ttstart [expr {$ttstart - 4}]
   if {$ttstart < 0} then  {
      set ttstart 0
   }
   ttredisplay
}

proc click-ttrestart {} {
   global ttstart
   set ttstart 0
   ttredisplay
}

proc click-ttrefresh {v args} {
   global debug afterid tensecs ttrefresh

   if {$debug} then {puts stderr "click-ttrefresh $v"}
   if {$afterid!=""} then {kill-after $afterid}
   set afterid ""
   if {$v==0} then {set ttrefresh 1; ttredo} else {set ttrefresh 0}
}

proc click-mdisplay {m args} {
   if {$m=="Redisplay"}                then {redisplayall}
   if {$m=="Disable Inactive Buttons"} then {:modify SERVER :inactive T; click-force 0}
   if {$m=="Enable All Buttons"}       then {:modify SERVER :inactive F; click-force 1}
}

proc click-redisplay {w} {
    set f [lowercase [first [split $w -]]]
    [set f]redisplay
}

set helptopic(DS) datastreamstatus
set helptopic(SS) streamstatus
set helptopic(FS) filestatus
set helptopic(TT) drivestatus

proc click-help {w} {
   global mytaskname helptopic

   set f [first [split $w -]]
   if {[info exists helptopic($f)]} then {
      send help help-on-topic $mytaskname $helptopic($f)
   } else {
      send help help-on-task $mytaskname
   }
}


#    set Tape Server transport to rpc if not supplied
if { ![info exists env(TCLTS)] } {set env(TCLTS) rpc}

proc click-server {args} {
  global env SERVER started
  global srv SRV

   switch [first $args] {
      RPC -
     {RPC Server}  {
       set env(TCLTS) "rpc"
       set srv eg
       set SRV EG
       loadsource tape.rpc
       set SERVER RPC
       if {$started} then {redisplayall}
       catch {:modify RESETDRIVE :inactive T}
     }
      Orbix -
     {Orbix Server}  {
       set env(TCLTS) "orbx"
       set srv orb
       set SRV ORB
       loadsource tape.orbx
       resumeTS
       set SERVER Orbix
       if {$started} then {redisplayall}
       catch {:modify RESETDRIVE :inactive F}
     }
      Claim -
     {Claim Orbix Server}  {
       set env(TCLTS) "orbx"
       set srv orb
       set SRV ORB
       loadsource tape.orbx
       claimTS
       set SERVER Orbix
       if {$started} then {redisplayall}
     }
      Free -
     {Free Orbix Server}  {
       set env(TCLTS) "orbx"
       set srv orb
       set SRV ORB
       loadsource tape.orbx
       freeTS
       set SERVER Orbix
     }
   }
}

if {$env(TCLTS) == "orbx"} then {click-server Orbix} else {click-server RPC}

proc click-resetdrive {args} {
#    reset Orbix Tape Drive Server (not Tape Manager)

   global srv server alias
   set z [catch {$srv "(tape :path '$server' :alias '$alias' !reset)" } m]
   if {$z!=0} then {
      display-$srv-error-report self $m "Attempting to reset drive."
      clear-footer
      error $m
   }
}

proc click arg {
   global debug openmode tsautomode
   if {$debug} then {puts stderr "click with argument \{$arg\}"}
   set f [first $arg]
   set c [second $arg]
   set w [third $arg]
   set v [fourth $arg]
   if {$c==":quit"} then {click-quit $f; return}

   switch -glob $w {
      *-REDISPLAY  {click-redisplay $w}

      FDLIST       {click-fdlist $c $v}
      UDLIST       {click-udlist $c $v}
      TSUDLIST     {click-tsudlist $c $v}

      DO-OPEN      {
          if {$openmode == 0} then {click-lblopen} else {click-unlblopen}
      }

      AUTOMODE     {set tsautomode $v}
      default {
         set p "click-[lowercase $w]"
         if {[info procs $p] == "$p"} then {$p $v} else {set-footer "no handler for widget $w"}
      }
   }
}

proc start args {
   global started alias drive curdrv freelist

   if {$started != 0} then {open-frame} else {make-frame}
   find-used-drives
   set curdrv ""
   set alias ""
   set drive [first $freelist]
   redisplayall
}

proc make-frame {} {
  global env started SERVER srv server aliases

     make_tpframe

     click-server $SERVER

     if {$env(TCLTS) == "orbx"} then {
        set z [catch {$srv "(tape :path '$server' :alias '[first $aliases]' ?fstatus)" } m]
        if {$z && $m==8} then {click-server Claim}
     }

     set started 1
}

proc click-quit {f} {close-frame $f}

proc finish args {destroy-frame}

proc kill-after {id} {after cancel $id}


proc make_tpframe {} {
  global env curdrv driveslen
  global SRV
  global negstreams egstreamlist tsautomode fopenoptions

   :frame [self frame-name] :label [self frame-title] :show-footer T  :resize F \
      :icon-image "@$env(ICONHOME)/default" \
      :icon-name "[self icon-title]"

      :panel TP-PANEL :layout L :padx 10 :pady 5
         :layout H; :row-gap 10

        :message M0 :label "Selected drive:"  :bold T
        :message TP-CURDRV :label $curdrv
        :hskip 30
        :message M1 :label State:  :bold T
        :message TP-STATE :label XXXXXXXXXXXX
        :next-row T

        :message M4 :label Drive: :valuex 12c :bold T
        :button CLAIM :label Claim
        :button FREE :label Free
        :next-row T

        :message M5 :label Volume: :valuex 12c :bold T
        :menubutton INIT :label Initialise :strings [list New Reinitialise Custom]
        :menubutton MMOUNT :label Mount :strings [list Identify Mount]
        :button DISMOUNT :label Dismount
        :next-row T

        :message M6 :label File: :valuex 12c :bold T
        :menubutton MOPEN :label Open :strings $fopenoptions
        :button CLOSE :label Close
        :next-row T

        :message M60 :label "" :valuex 12c :bold T
        :button REWIND :label "Rewind Tape" :inactive T
        :button WRITETM    :label "Write Tape Mark" :inactive T
        :button SKIP   :label "Skip" :inactive T :command make_skpanel

    :panel TP-LIST1 :layout L :pady 5
       :layout V
        :message TP-M1 :label "Drives Available"
        :list FDLIST  :rows 8 :cols $driveslen :exclusive T

    :panel TP-LIST2 :layout L :padx 5 :pady 5
       :layout V
        :message TP-M2 :label "Drives Claimed"
        :list UDLIST  :rows 8 :cols $driveslen :exclusive T

      set egstreamlist {}
      for {set i 1} {$i <= $negstreams} {incr i} {lappend egstreamlist $i}

   :panel TS-PANEL :layout B :padx 5 :pady 5 :after FOOTER
       :layout H; :row-gap 10; :col-gap 10

       :number NUMSTRMS :label "Data Streams" :min 1 :max 4 :value $negstreams
       :choice EGSTREAM :label "Data Stream:" :valuex 15c  \
           :exclusive T :strings $egstreamlist :value 0
       :hskip 80
       :choice EGSTATE :label "Server state:"  \
           :exclusive T :strings [list Stop Go Test] :value 0
       :next-row T
       :text STRINGS :label "Association:"  :w 42  :valuex 15c
       :button CLEAR :label "Clear"
       :checkbox AUTOMODE :label "Auto" :value $tsautomode
       :list TSUDLIST :rows 8 :cols 12 :exclusive T
       :next-row T
        :vskip -100
        :hskip 100
        :button ASSOC  :label "Associate"
        :button DISSOC :label "Dissociate"
        :col-gap 10
        :choice PLUS :label "Join next drive with:" :exclusive T \
             :strings [list " " "+"] :value 0
        :layout V; :next-row T; :row-gap 5
        :message M2 :label "Current Data Stream Associations" :bold T
        for {set i 1} {$i <= $negstreams} {incr i} {
           :text ASSOCS$i :label "Stream $i" :valuex 15c :w 42
        }

    :panel TP-BUTTONS :layout B :padx 10 :after FOOTER
        :layout H; :col-gap 5; :row-gap 10
        :menubutton MDISPLAY :label Redisplay \
            :strings [list Redisplay "Enable All Buttons" "Disable Inactive Buttons"]
        :button CHANGER  :label Changer :inactive T
           :menubutton STATUS :label Status :strings [list Summary File Drive Streams]
            :col-gap 0
        :menubutton SERVER :label Server :inactive T  \
           :strings [list "RPC Server" "Orbix Server" "Claim Orbix Server" "Free Orbix Server"]
        :button RESETDRIVE :label "Reset Drive" :inactive T
        :col-gap 10
        :button HELP :command "click-help TP-HELP"

   :panel DIALOG :layout B :padx 5 :pady 0 :w 100 :h 120 :after TS-PANEL
       [:path DIALOG] configure -relief raised -bd 1

    :show T
}

proc empty_dialog_panel {} {:modify DIALOG :empty T}

proc make_dialog_panel {} {
   global curdrv devstates alias

   empty_dialog_panel
   update

   if {$alias==""} then {
      set state 0
   } else {
      set st [query-file-status $alias]
      set state [second $st]
      set v [lindex $st 5]
      set f [first [split [lindex $st 6] /]]
   }

   :layout V; :row-gap 0
   if {$state != 0} then {
      :message D10 :label "You can select a tape from the \"Drives Claimed\" list at any time"
      :message D20 :label "The currently selected drive is $curdrv and it is in the state [lindex $devstates $state]"
   } else {
      :message D10 :label "You must first select a drive from the \"Drives Available\" list and then \"Claim\""
   }

   switch $state {
     0  {
        }
     1  {
          :message D30 :label "You should now put a tape in the drive and then \"Initialise\" or \"Mount\""
        }
     3  {
          :message D30 :label "You can now either create a file on the tape using \"Open\" or remove the tape using \"Dismount\". Volume=$v."
        }
     5  {
          :message D30 :label "The tape is ready to write data. Volume=$v; file=$f. If you want to remove the tape first use \"Close\""
        }
     2  -
     4  -
     6  -
     7  -
     8  -
     9  -
     12 -
     18 -
     20 -
     22 -
     23 {
          :message D30 :label "The drive is currently busy - wait until the operation has completed"
     }
     10 {
          :message D30 :label "The drive is currently writing data to the tape"
     }
     16 {
          :message D30 :label "There has been a tape error - You will need to change to a new tape - First use \"Close\" or \"Reset Drive\""
     }
     default {
          :message D30 :label "The drive is not in an expected state - if this does not change soon you may need assistance"
     }
   }
   :message D99 :label "For more information click on the \"Help\" button or use the F1 key"
}

proc make_copanel {} {
  global curdrv amodenameslist labelnameslist fileslen filename amodeindex labelindex rlen blen

     empty_dialog_panel

        :layout H; :row-gap 5; :col-gap 0
        :message D1 :label "Open File on drive " :bold T
        :message DRIVE :label $curdrv :bold T
        :col-gap 10
        :hskip 100
        :button OPRESET :label "Set Defaults"
        :next-row T
        :text FILE :label Filename: :w $fileslen :valuex 15c :value $filename
        :choice-stack MODE :label "Access mode:" :strings $amodenameslist :value $amodeindex
        :hskip 50
        :choice-stack LABEL :label "Labelling:" :strings $labelnameslist :value $labelindex
        :next-row T
        :number RLEN :label "Record length:" :w 6 :valuex 15c :value $rlen :max 65536
        :number BLEN :label "Block length:"  :w 6 :valuex 15c :value $blen :max 65536
        :hskip 100
        :button CUSTOPEN :label Execute
}

proc make_dopanel {mode} {
 global curdrv filename fileslen

     empty_dialog_panel

        :layout H; :row-gap 5
        :message D1 :label "Open File on drive " :bold T
        :message DRIVE :label $curdrv :bold T
        :next-row T
        if {$mode == 0} then {
           :text FILE :label Filename: :w $fileslen :valuex 15c :value $filename
        } else {
           :message D2 :label "Tape is unlabelled"
        }
        :hskip 100
        :button DO-OPEN :label Execute
}

proc make_nipanel {} {
  global curdrv volume automount volslen
  global initvols alias

     empty_dialog_panel

   set volume $initvols($alias)

        :layout H; :row-gap 5
        :message D1 :label "Initialise a new tape on drive " :bold T
        :message DRIVE :label $curdrv :bold T
        :col-gap 5
        :next-row T
        :text INITVOL :label "New volume name:" :w $volslen :valuex 20c :value $volume
        :next-row T
        :checkbox AUTOMOUNT :label Mount: :value $automount :valuex 20c
        :layout H; :hskip 100
        :button NEWINIT :label Execute

   set-footer "Initialise a new tape"
}

proc make_ripanel {} {
  global curdrv volume oldvol automount volslen
  global initvols alias 

     empty_dialog_panel

   set volume $initvols($alias)

        :layout H; :row-gap 5
        :message D1 :label "Re-initialise existing tape on drive " :bold T
        :message DRIVE :label $curdrv :bold T
        :col-gap 5
        :next-row T
        :text INITVOL :label "New volume:" :w $volslen :valuex 20c :value $volume
        :text OLDVOL  :label "Old volume:" :w $volslen             :value $oldvol
        :next-row T
        :checkbox AUTOMOUNT :label Mount: :value $automount :valuex 20c
        :hskip 100
        :button REINIT :label Execute

   set-footer "Reinitialise a tape"
}

proc make_cipanel {} {
  global curdrv volume vlabelnameslist densitynameslist oldvol automount volslen
  global initvols alias densityindex vlabelindex

     empty_dialog_panel

   set volume $initvols($alias)

        :layout H; :row-gap 5
        :message D1 :label "Initialise tape on drive " :bold T
        :message DRIVE :label $curdrv :bold T
        :col-gap 5
        :hskip 100
        :button CIRESET :label "set defaults"
        :next-row T
        :text INITVOL :label "New volume:" :w $volslen :valuex 15c :value $volume
        :text OLDVOL:label "Old volume:" :w $volslen :valuex 15c :value $oldvol
        :next-row T
        :choice-stack VLABEL :label "Labelling:" :valuex 15c \
             :strings $vlabelnameslist :value $vlabelindex
        :choice-stack DENSITY :label Density: :strings $densitynameslist :value $densityindex
        :checkbox AUTOMOUNT :label Mount: :value $automount 
        :hskip 100
        :button CSTINIT :label Execute

   set-footer "Custom initialise"
}

proc make_mnpanel {} {
  global curdrv volslen volume

     empty_dialog_panel

        :layout H; :row-gap 5
        :message D1 :label "Mount Volume on drive " :bold T
        :message DRIVE :label $curdrv :bold T
        :col-gap 5
        :next-row T
        :text VOLUME :label "Volume name:" :w $volslen :valuex 15c :value $volume
        :hskip 100
        :button MOUNT :label Execute

   set-footer "Mount a tape"
}

proc make_skpanel {args} {
 global curdrv alias skcount skmode skdirctn

     empty_dialog_panel

        :layout H; :row-gap 5
        :message D1 :label "Skip Blocks(s) or Tape Marks(s) on drive " :bold T
        :message DRIVE :label $curdrv :bold T
        :col-gap 5
        :next-row T
        :choice SKDIRCTN :label Direction: :valuex 12c \
                :strings [list Backwards Forwards] :value $skdirctn
        :choice SKMODE :label Skip: :valuex 12c \
                :strings [list "Tape Marks" "Data Blocks"] :value $skmode
        :next-row T
        :number SKCOUNT :label "Number:" :w 4 :valuex 12c :value $skcount :min 1 :max 256
        :hskip 100
        :button SKIP :label Execute

   if {[lindex [query-file-status $alias] 8] == 8} then {
       :modify SKIP :inactive F
   } else {
       :modify SKIP :inactive T
   }

   set-footer "Skip Block(s) or Tape Marks(s)"
}

set fsframe  "[self frame-name].fs"

proc make_fsframe {} {
  global env driveslen volslen fileslen
  global fsframe

set offset1 22c
set offset2 14c

  :frame $fsframe :label "Current File Status" :show-footer T :resize F :transient T

    :panel FS-PANEL1 :layout L :padx 10
     :layout V; :row-gap 5
     :vskip 5

        :text FS-LAST :label "Last procedure status:" :w 36 :read-only T :valuex $offset1
        :text FS-STATE :label "Current stream state:" :w 36 :read-only T :valuex $offset1
        :text FS-DEVICE :label "Requested Device:" :w 20 :read-only T :valuex $offset1
        :text FS-RDEVICE :label "Allocated Device:" :w 20 :read-only T  :valuex $offset1
        :text FS-VOLUME :label Volume: :w 20 :read-only T :valuex $offset1
        :text FS-FILE :label File: :w 20 :read-only T  :valuex $offset1
        :text FS-MODE :label Mode: :w 20 :read-only T  :valuex $offset1
        :text FS-LABELS :label Labels: :w 20 :read-only T :valuex $offset1

    :panel FS-PANEL2 :layout L :padx 10
     :layout V; :row-gap 5

      :vskip 20
        :text FS-RLEN :label "Record Length:" :w 12 :read-only T :valuex $offset2
        :text FS-BLEN  :label "Block Length:" :w 12 :read-only T  :valuex $offset2
        :text FS-DLEN :label "Data Length:" :w 12 :read-only T :valuex $offset2
      :vskip 20
        :text FS-BLOCK :label "Block count:" :w 12 :read-only T :valuex $offset2
        :text FS-BYTE :label "Kbyte count:" :w 12 :read-only T  :valuex $offset2
        :text FS-RATE :label Kbytes/sec: :w 12 :read-only T  :valuex $offset2

    :panel FS-PANEL3 :layout B :padx 10 :after FOOTER
     :layout H; :row-gap 5

        :text FS-INFO :label Information: :w 80 :read-only T
        :next-row T

        :button FS-REDISPLAY :label Redisplay
        :hskip 80
        :button HELP :command "click-help FS-HELP"

     :show T
}

set dsframe  "[self frame-name].ds"

proc make_dsframe {} {
  global env curdrv alias driveslen
  global dsframe

  :frame $dsframe :label "Current Drive Status" :show-footer T :resize F :transient T

    :panel DS-PANEL :layout L :padx 10 :pady 10
     :layout V; :row-gap 5

        set offset 15c
        :text DS-DRIVE :label Drive: :w $driveslen :valuex $offset :value $curdrv :read-only T
        :text DS-TYPE :label "Device Type:" :w 12 :valuex $offset :read-only T
        :text DS-TLEN :label "Tape Length:" :w 12 :valuex $offset :read-only T
        :text DS-USED :label "Tape Used:" :w 12 :valuex $offset :read-only T
        :text DS-REM :label  "Tape Left:" :w 12 :valuex $offset :read-only T
        :text DS-ERR :label  "Tape Errors:" :w 12 :valuex $offset :read-only T
        :text DS-ERRPC :label "Tape % Errors:" :w 12 :valuex $offset :read-only T

        :button DS-REDISPLAY  :label Redisplay
        :layout H
        :button DS-DETAIL  :label "Details" :command  show_device_sense
        :button HELP :command "click-help DS-HELP"

    :show T
}

set dsenframe  "[self frame-name].dsen"

proc show_device_sense {} {
  global env srv alias DLTInfo DLTmedia DLTdensity
  global dsenframe

  if {$alias == ""} then {return}

  :frame $dsenframe :label "Current Drive Sense Information" :show-footer T :resize F :transient T

    :panel DSEN-PANEL :layout L :padx 10 :pady 10
     :layout V; :row-gap 5

       :galley DSEN-INFO :rows 20 :cols 90

        if {[string toupper [string range [lindex [query-device-status $alias] 0] 0 2]] == "DLT"} then {
            if {$env(TCLTS) == "rpc"} then {
               set DLTsense [lrange $DLTInfo 9 38]
               set DLTinquiry [lrange $DLTInfo 69 124]
               set DLTmode [lrange $DLTInfo 45 56]
            } else {
               set DLTsense [lrange $DLTInfo 9 38]
               set DLTinquiry [lrange $DLTInfo 66 124]
               set DLTmode [lrange $DLTInfo 42 53]
            }

            :modify DSEN-INFO :strings "Sense Information :-\n$DLTsense\n\n"
            set i [expr 0x[lindex $DLTsense 2] &  0x80]
            if {$i == 1} then {:modify DSEN-INFO :strings "File Mark read by last command\n"}
            set i [expr 0x[lindex $DLTsense 2] &  0x40]
            if {$i == 1} then {:modify DSEN-INFO :strings "End of Medium found\n"}
            set i [expr 0x[lindex $DLTsense 2] &  0x20]
            if {$i == 1} then {:modify DSEN-INFO :strings "Incorrect Length Indicator set\n"}
            set i [expr 0x[lindex $DLTsense 0] &  0x80]
            set info "[lindex $DLTsense 3] [lindex $DLTsense 4] [lindex $DLTsense 5] [lindex $DLTsense 6]"
            if {$i == 1} then {:modify DSEN-INFO :strings "Information bytes :- $info\n"}
            :modify DSEN-INFO :strings "Sense Key :- [expr 0x[lindex $DLTsense 2] & 15]\n"
            :modify DSEN-INFO :strings "Additional Sense Code :- [lindex $DLTsense 12]\n"
            :modify DSEN-INFO :strings "Additional Sense Code Qualifier :- [lindex $DLTsense 13]\n"
            :modify DSEN-INFO :strings "Internal Status Code :- 0x[lindex $DLTsense 18]\n"
            set i [expr 0x[lindex $DLTsense 19][lindex $DLTsense 20]]
            :modify DSEN-INFO :strings "Tape Motion (hours) :- $i\n"
            set i [expr 0x[lindex $DLTsense 21][lindex $DLTsense 22][lindex $DLTsense 23][lindex $DLTsense 24]]
            :modify DSEN-INFO :strings "Power On (hours) :- $i\n"
            set i [expr 0x[lindex $DLTsense 25][lindex $DLTsense 26][lindex $DLTsense 27][lindex $DLTsense 28]]
            :modify DSEN-INFO :strings "Tape Remaining (4 KB) :- $i\n"

            :modify DSEN-INFO :strings "\n\nIn the event of repeated drive errors please give this information to your support group\n\n"

            :modify DSEN-INFO :strings "Mode Information :-\n$DLTmode\n\n"
            set i 0x[lindex $DLTmode 1]
            if {[info exists DLTmedia($i)]} then {set m $DLTmedia($i)} else {set m $DLTmedia(0x00)}
            :modify DSEN-INFO :strings "Media Type :- $i: $m\n"
            set i [expr 0x[lindex $DLTmode 2] &  0x80]
            if {$i == 1} then {
               :modify DSEN-INFO :strings "Tape is write protected\n"
            } else {
               :modify DSEN-INFO :strings "Tape is write enabled\n"
            }
            set i 0x[lindex $DLTmode 4]
            if {[info exists DLTdensity($i)]} then {set m $DLTdensity($i)} else {set m $DLTdensity(0x00)}
            :modify DSEN-INFO :strings "Density Code :- $i - $m\n\n\n"

            :modify DSEN-INFO :strings "Inquiry Information :-\n[lrange $DLTinquiry 0 7]\n[lrange $DLTinquiry 8 15]\n[lrange $DLTinquiry 16 31]\n[lrange $DLTinquiry 32 35]\n[lrange $DLTinquiry 36 55]\n"
 
            :modify DSEN-INFO :see 0
        } else {
            :modify DSEN-INFO :strings "Currently only available for DLT tapes"
        }

    :show T
}

set ttframe  "[self frame-name].tt"

proc make_ttframe {} {
  global env tblrows driveslen volslen fileslen
  global ttframe
  global SRV

  :frame $ttframe :label "Tape Drive Summary" :show-footer T :resize F :transient T

    :panel TT-PANEL1 :layout L :padx 10
     :layout H; :row-gap 5
     :vskip 5

        :message M1 :label Drive          :bold T  :valuex 9c
        :message M2 :label Volume         :bold T  :valuex [expr $driveslen]c
        :message M3 :label File           :bold T  :valuex [expr $volslen + 1]c
        :message M4 :label State          :bold T  :valuex [expr $fileslen + 1]c
        :message M5 :label Kbytes/s       :bold T  :valuex 10c
        :message M6 :label "Tape Used"    :bold T  :valuex 12c
        :message M7 :label "Tape Errors"  :bold T  :valuex 15c
        :message M8 :label "Time Left"    :bold T  :valuex 12c

        :next-row T

      for {set i 0} {$i<$tblrows} {incr i} {

         :text TT-DRIVE$i   :w $driveslen :read-only T :label ""
         :text TT-VOLUME$i  :w $volslen :read-only T :label ""
         :text TT-FILE$i    :w $fileslen :read-only T :label ""
         :text TT-STATE$i   :w 9 :read-only T :label ""
         :text TT-RATE$i    :w 8 :read-only T :label ""
            :hskip 10
            :gauge TT-USED$i   :value 0 :min 0 :max 0 :label ""
            :gauge TT-ERRS$i   :value 0 :min 0 :max 0 :label "   "
            :gauge TT-TLEFT$i  :value 0 :min 0 :max 0 :label "   "
         :next-row T
      }

         :row-gap -30; :next-row T
      :hskip 40
         :choice TTREFRESH  :label Refresh: :strings [list On Off] :exclusive T :value 0
    :hskip 200
         :choice TT-EGSTATE :label "Tape server state:" :valuex 20c \
             :exclusive T :strings [list Stopped Going Test] :value 0 :deaf T

         :row-gap 10; :next-row T; :col-gap 10
         :button TT-REDISPLAY  :label Redisplay
         :menubutton TTDRIVES :label "More drives" \
             :strings [list "Start of List" "Next" "Previous" "Next 4" "Previous 4"]
         :hskip 80
         :button HELP :command "click-help TT-HELP"

    :show T
}

set ssframe  "[self frame-name].ss"

proc make-ssframe {} {
  global env ssframe negstreams

 :frame $ssframe :label "Data Stream Summary" :show-footer F \
      :icon-image "@$env(ICONHOME)/default" \
      :icon-name "[self icon-title]"

 wm protocol $ssframe \
    WM_DELETE_WINDOW {global ssvisible; set ssvisible 0; destroy $ssframe}

   :panel SS-PANEL :layout T :padx 10 :pady 5
        :layout H; :row-gap 10

        :message SS-LABEL0 :label "Stream:"  :bold T  :valuex 10c
        :message SS-LABEL1  :label 1 :bold T  :valuex 5c
        for {set i 2} {$i <= $negstreams} {incr i} {
          :message SS-LABEL$i  :label $i :bold T  :valuex 18c
        }

        :next-row T
        for {set i 1} {$i <= $negstreams} {incr i} {
           :text SS-BLOCK$i :label "Blocks:" :w 8 :read-only T \
               :valuex 10c :value 0
        }

        :next-row T
        for {set i 1} {$i <= $negstreams} {incr i} {
           :text SS-BYTE$i :label "Kbytes:" :w 8 :read-only T  \
               :valuex 10c :value 0
        }

        :next-row T
        for {set i 1} {$i <= $negstreams} {incr i} {
           :text SS-RATE$i :label "Kbytes/s:" :w 8 :read-only T  \
               :valuex 10c :value 0
        }

        :next-row T; :layout H
        :button SS-REDISPLAY  :label "Redisplay"
        :hskip 80
        :button HELP :command "click-help SS-HELP"

    :show T
}

#     repaint procedures

proc repaint-dflt {} {
  global forcebuttons
  global nivisible civisible rivisible mnvisible dovisible covisible skvisible
   if {$forcebuttons==0} then {set inact "T"}  else {set inact "F"}

             :modify CLAIM  :inactive $inact
             :modify FREE :inactive $inact
             :modify INIT :inactive $inact
             :modify MMOUNT :inactive $inact
             :modify DISMOUNT :inactive $inact
             :modify MOPEN :inactive $inact
             :modify REWIND :inactive $inact
             :modify WRITETM :inactive $inact
             :modify SKIP :inactive $inact
             :modify CLOSE :inactive $inact

  update

       make_dialog_panel

  update
}

proc repaint {} {
   global env debug alias curdrv allocated devstates \
          unallocated allocated mounting mounted opening open closing \
          executing initialising identifying putting \
          moving twelve thirteen fourteen fifteen errorstate 

if {$env(TCLTS) == "orbx"} then {
    global rewinding skipping importing writingfm test  waitforhdr writinghdr
}

   if {$alias==""} then {
      set state 0
   } else {
      set state [second [query-file-status $alias]]
   }

   if {$debug} then {puts stdout "repaint: state [lindex $devstates $state] ($state)"}

   :modify TP-CURDRV :label $curdrv
   :modify TP-STATE  :label [lindex $devstates $state]

   global forcebuttons
   if {$forcebuttons==0} then {set inact "T"}  else {set inact "F"}

   repaint-dflt

   switch $state {

     0  {:modify CLAIM  :inactive F}
     1  {
             :modify FREE  :inactive F
             :modify INIT  :inactive F
             :modify MMOUNT  :inactive F
             :modify DISMOUNT :inactive F
        }
     2  {
             :modify FREE :inactive F
        }
     3  {
             :modify INIT :inactive F
             :modify DISMOUNT  :inactive F
             :modify MOPEN  :inactive F
             :modify REWIND :inactive F
        }
     5  {
             if {[lindex [query-file-status $alias] 8] == 8} then {
                :modify WRITETM :inactive F
                :modify SKIP :inactive F
             }
             :modify CLOSE  :inactive F
         }
     10  {:modify CLOSE  :inactive F}
     16  {
             :modify CLOSE    :inactive F
             :modify DISMOUNT  :inactive F
         }

     4  -
     6  -
     7  -
     8  -
     9  -
    11  -
    12  -
    13  -
    14  -
    15   {}
    }
}

proc ttrepaint {} {
   global tblrows ttstart aliaslist usedlist egstatenames devstates smodeslist \
           unallocated allocated mounting mounted opening open \
           closing executing initialising identifying putting \
           moving twelve thirteen fourteen fifteen error
   global egstate SRV units

   set egstate [read-egstate]
   :modify TT-EGSTATE :value [expr {$egstate -1}]

   for {set i 0} {$i<$tblrows} {incr i} {
      set a [lindex $aliaslist [expr {$i + $ttstart}]]
      if {$a != ""} then  {
         set st [query-file-status $a]
         set pstate [second $st]
         set widget [format "TT-DRIVE%d" $i]
         set dr [lindex $usedlist [expr {$i + $ttstart}]]
         :modify $widget :value $dr
         set widget [format "TT-VOLUME%d" $i]
         if {$pstate>$allocated && $pstate<$twelve} then {
            set v [lindex $st 5]
            } else {
            set v ""
            }
         :modify $widget :value $v
         set widget [format "TT-FILE%d" $i]
         if {$pstate>$mounted && $pstate<$twelve} then {
            set v [first [split [lindex $st 6] /]]
            } else {
            set v ""
            }
         :modify $widget :value $v
         set widget [format "TT-STATE%d" $i]
         :modify $widget :value "[lindex $devstates $pstate]"
         set widget [format "TT-RATE%d" $i]
         if {$pstate>$opening && $pstate<$twelve} then {
            set v [expr [lindex $st 17]/$units]
            } else {
            set v ""
            }
         :modify $widget :value $v
         set st [query-device-status $a]
         set widget [format "TT-USED%d" $i]
         if {$pstate>$mounting && $pstate<$twelve} then {
            set v1 [second $st]
            set v2 [expr [second $st]-[third $st]]
         } else {
            set v1 0
            set v2 0
         }
         :modify $widget :max $v1 :value $v2
         set widget [format "TT-ERRS%d" $i]
         if {$pstate>$mounting && $pstate<$twelve} then {
            set v2 [fifth $st]
            if {$v2>10} then {
            set v1 100
            set v3 "10 %"
            } else {
            set v1 10
            set v3 "1 %"
            }
            } else {
            set v1 0; set v2 0; set v3 0
            }
         :modify $widget :max $v1 :value $v2 :max-tick-label $v3
         set widget [format "TT-TLEFT%d" $i]
         if {$pstate>$mounting && $pstate<$twelve} then {
            if {$v>0} then {
            set v2 [expr [third $st]/$v]
            } else {
            set v2 180000
            }
            if {$v2>12000} then {
            set v1 180000
            set v3 "50 hrs"
            } else {
               if {$v2>1000} then {
               set v1 12000
               set v3 "200 mins"
               } else {
               set v1 1000
               set v3 "1000 secs"
               }
            }
          } else {
            set v1 0; set v2 0; set v3 0
          }
         :modify $widget :max $v1 :value $v2 :max-tick-label $v3

      }  else  {
         ttclearaline $i
      }
   }
}


proc ttclearaline l  {
  global SRV
   :modify TT-DRIVE$l  :value ""
   :modify TT-VOLUME$l :value ""
   :modify TT-FILE$l   :value ""
   :modify TT-STATE$l :value ""
   :modify TT-RATE$l   :value ""
   :modify TT-USED$l   :value 0 :max 0
   :modify TT-ERRS$l  :value 0 :max 0
   :modify TT-TLEFT$l  :value 0 :max 0

}

proc dsrepaint {} {
   global  curdrv alias
   :modify DS-DRIVE :value $curdrv
   if {$alias==""} then {
       :modify DS-TYPE  :value ""
       :modify DS-TLEN  :value ""
       :modify DS-USED  :value ""
       :modify DS-REM   :value ""
       :modify DS-ERR   :value ""
       :modify DS-ERRPC :value ""
      }  else  {
        set st [query-device-status $alias]
        :modify DS-TYPE :value  [first $st]
        :modify DS-TLEN :value  [second $st]
        :modify DS-USED :value  [expr [second $st]-[third $st]]
        :modify DS-REM  :value  [third $st]
        :modify DS-ERR  :value  [fourth $st]
        :modify DS-ERRPC :value [expr [fifth $st]*0.1]
      }
}

proc fsrepaint {} {
   global SERVER alias devstates procstatuses amodelist amodenameslist labellist labelnameslist units
   if {$alias==""} then {
     :modify  FS-LAST    :value ""
     :modify  FS-STATE   :value ""
     :modify  FS-INFO    :value ""
     :modify  FS-DEVICE  :value ""
     :modify  FS-RDEVICE :value ""
     :modify  FS-VOLUME  :value ""
     :modify  FS-FILE    :value ""
     :modify  FS-MODE    :value ""
     :modify  FS-LABELS  :value ""
     :modify  FS-RLEN    :value ""
     :modify  FS-BLEN    :value ""
     :modify  FS-DLEN    :value ""
     :modify  FS-BLOCK   :value ""
     :modify  FS-BYTE    :value ""
     :modify  FS-RATE    :value ""
   }  else  {
        set st [query-file-status $alias]
        set status [first $st]
        set statusx 0x[format %x $status]
        set procstatus [lindex $procstatuses $status]
        if {$procstatus != ""} then {
            :modify FS-LAST   :value "$procstatus"
        } else {
            :modify FS-LAST   :value [format "%s (%s)" $statusx $status]
        }
     :modify FS-STATE  :value [lindex $devstates [second $st]]
     :modify FS-INFO   :value [third $st]
     :modify FS-DEVICE :value [fourth $st]
     :modify FS-RDEVICE :value [lindex $st 4]
     :modify FS-VOLUME :value [lindex $st 5]
     :modify FS-FILE   :value [first [split [lindex $st 6] /]]
     set v [lsearch $amodelist [lindex $st 7]]
     :modify FS-MODE   :value [lindex $amodenameslist $v]
     set v [lsearch $labellist [lindex $st 8]]
     :modify FS-LABELS :value [lindex $labelnameslist $v]
     :modify FS-RLEN   :value [lindex $st 9]
     :modify FS-BLEN   :value [lindex $st 10]
     :modify FS-DLEN   :value [lindex $st 11]
     :modify FS-BLOCK  :value [lindex $st 15]
     :modify FS-BYTE   :value [expr [lindex $st 15] * ([lindex $st 10]/1024)]
     :modify FS-RATE   :value [expr [lindex $st 17]/$units]
   }
}

proc drivesrepaint {} {
   global curdrv freelist usedlist
   find-used-drives
   set freelist [concat "TAPE" [find-fdrives [read-drives]]]

   :modify FDLIST :append F :strings  $freelist
   :modify FDLIST :select  [lsearch $freelist $curdrv]
   :modify UDLIST :append F :strings  $usedlist
   :modify UDLIST :select  [lsearch $usedlist $curdrv]
   :modify TSUDLIST :append F :strings  $usedlist
   :modify TSUDLIST :select -1
}


# DONT DISTINGUISH BETWEEN TAPE DEVICE TYPES
proc drivesrepaint {} {
   global curdrv freelist usedlist
   find-used-drives
   set freelist [find-fdrives [read-drives]]

   :modify FDLIST :append F :strings  $freelist
   :modify FDLIST :select  [lsearch $freelist $curdrv]
   :modify UDLIST :append F :strings  $usedlist
   :modify UDLIST :select  [lsearch $usedlist $curdrv]
   :modify TSUDLIST :append F :strings  $usedlist
   :modify TSUDLIST :select -1
}


#    for use by the rates and monitor task

proc update-information {} {
   global tblrows aliaslist ttstart usedlist \
           unallocated allocated mounting mounted opening open \
           closing executing initialising identifying putting \
           moving twelve thirteen fourteen fifteen error
   global taperates tapelengths tapeused units

   set taperates {}
   set tapelengths {}
   set tapeused {}

   for {set i 0} {$i<$tblrows} {incr i} {
      set a [lindex $aliaslist [expr {$i + $ttstart}]]
      if {$a != ""} then  {
         set st [query-file-status $a]
         set pstate [second $st]
         if {$pstate>$opening && $pstate<$twelve} then {
            set v [expr [lindex $st 17]/$units]
         } else {
            set v 0
         }
         lappend taperates $v
         set st [query-device-status $a]
          if {$pstate>$mounting && $pstate<$twelve} then {
            set v1 [second $st]
            set v2 [expr [second $st]-[third $st]]
         } else {
            set v1 0; set v2 0
         }
         lappend tapelengths $v1
         lappend tapeused $v2
      }
   }
}
